/*
 * main.c
 *
 * The purpose of this code is to read data from hall effect sensors, blink an
 * LED, read data from 16 12-bit ADC channels, and send data over serial to a computer. The behavior of the code is modified by a configuration string sent by
 * a computer via serial to USB.
 *
 *
 *  Created on: Nov 8, 2012
 *      Author: Michael Kilzer (mjk277)
 */


// Files to Include
#include "LPC17xx.h"
#include <cr_section_macros.h>
#include <NXP/crp.h>
#include <stdio.h>
#include <stdlib.h>
#include "uart2.h"
#include "spi.h"
#include <string.h>

// Definitions
// #define CLKFREQ 100000000

#define CLKDIV0 0b01
#define MODE0 0b11
#define PRESCALER0 1000
#define TIME_INT0 16666

// 16666 = 500ms


#define CLKDIV1 0b01
#define MODE1 0b11
#define PRESCALER1 1000
#define TIME_INT1  3333


#define EINT0_CLR (1 << 0)
#define EINT1_CLR (1 << 1)
#define EINT2_CLR (1 << 2)
#define EINT3_CLR (1 << 3)


#define EINT0_EDGE	(1 << 0)
#define EINT1_EDGE	(1 << 1)
#define EINT2_EDGE  (1 << 2)
#define EINT3_EDGE  (1 << 3)

// PWM Values
#define PWM_CLKDIV 0b11 // Divide by 1
#define PWM_MODE 0b01 // Set pin to PWM
#define MAXPWM 8754 //ticks
#define MINPWM 3754 //ticks


// Variable to store CRP value in. Will be placed automatically
// by the linker when "Enable Code Read Protect" selected.
// See crp.h header for more information
__CRP const unsigned int CRP_WORD = CRP_NO_CRP ;

// Variables

typedef struct struct_Datum
{
    uint16_t rpm0, rpm1, rpm2, rpm3,
        adc0, adc1, adc2, adc3, adc4, adc5,
        adc6, adc7, adc8, adc9, adc10, adc11,
        adc12, adc13, adc14, adc15;

} Datum;

volatile uint8_t rpm_flag, dynorun_flag, timingFlag;
volatile uint32_t calc_count1, count1, count0, count2, count3;
volatile uint32_t calc_count0, calc_count2, calc_count3;

uint32_t ledstate, time, pwmduty, i, runs, duration, j;
uint16_t ADCdata[16];


uint8_t PWM_flag;

// Used for sending and receiving data from the ADCs.
uint8_t rec[8][2];
uint8_t send[8][2] = {{0b10000011, 0b00010000},
					 {0b10000111, 0b00010000},
					 {0b10001011, 0b00010000},
					 {0b10001111, 0b00010000},
					 {0b10010011, 0b00010000},
					 {0b10010111, 0b00010000},
					 {0b10011011, 0b00010000},
					 {0b10011111, 0b00010000}};
uint8_t datastr[90];

Datum DynoData;
uint8_t  r, checksum, checksumByteOne, checksumByteTwo;

uint8_t samplePeriod[3], ADCStatus[7], eintStatus[3];
uint8_t rcvData[40], rcvDataFlag;

uint8_t configStringFlag, configDataReady, configStatus;
uint16_t samplePeriodInt, ADCStatusConv, eintStatusConv;

// Prototypes
void timer_init(void);
void eint_init(void);
void pwm_init(void);
void display_data(void);
void calc_rpm_data(void);
void leds_invert(void);
void sampleADC(void);
void DAQ_Config(uint8_t* config_str);
void receiveConfigData(char c);
void asciiSensorStatus(uint16_t eintStatusConv, uint32_t ADCStatusConv, uint8_t* datastr);



// Used to initialize both Timer0 and Timer1.
void timer_init(void)
{
	
	// Timer0
	LPC_TIM0->MR0 = TIME_INT0; // Interrupt Generated when MR0 is this value.
	LPC_TIM0->MCR = MODE0;     // Interrupt and Reset on MR0 match.
	LPC_SC->PCLKSEL0 &= ~(0b11 << 2);  //clear bits
	LPC_SC->PCLKSEL0 |= (CLKDIV0 << 2);  //set bits
	LPC_TIM0->PR = PRESCALER0; // Prescaler
	NVIC_EnableIRQ(TIMER0_IRQn); // Enable Interrupt
	LPC_TIM0->TCR = 1;           // Enable Counter

	// Timer1
	LPC_TIM1->MR0 = (int)(33.33*samplePeriodInt);
	LPC_TIM1->MCR = MODE1;
	LPC_SC->PCLKSEL0 &= ~(0b11 << 4);  //clear bits
	LPC_SC->PCLKSEL0 |= (CLKDIV1 << 4);  //set bits
	LPC_TIM1->PR = PRESCALER1;
	NVIC_EnableIRQ(TIMER1_IRQn);
	NVIC_SetPriority(TIMER1_IRQn, 1);
	LPC_TIM1->TCR = 1;

}


void eint_init(){

	// EINT0 Configurations
	//set the interrupt function to use the supplied callback
	// Set PINSEL4 [21:20] = 01 for P2.10 as EINT0
	LPC_PINCON->PINSEL4 &= ~(0b11 << 20);
	LPC_PINCON->PINSEL4 |= (0b01 << 20);
	//always clear the flag before initializing
	LPC_SC->EXTINT = EINT0_CLR;
	// EINT0 user selected trigger
	LPC_SC->EXTMODE |= EINT0_EDGE;
	// EINT0 is falling edge by default
	LPC_SC->EXTPOLAR &= ~(1 <<0);
	// Enable External Interrupt 0
	NVIC_EnableIRQ(EINT0_IRQn);
	NVIC_SetPriority(EINT0_IRQn, 0);


	// EINT1 Configurations
	//set the interrupt function to use the supplied callback
	// Set PINSEL4 [23:22] = 01 for P2.11 as EINT1
	LPC_PINCON->PINSEL4 &= ~(0b11 << 22);
	LPC_PINCON->PINSEL4 |= (0b01 << 22);
	//always clear the flag before initializing
	LPC_SC->EXTINT = EINT1_CLR;
	// EINT1 user selected trigger
	LPC_SC->EXTMODE |= EINT1_EDGE;
	// EINT1 is falling edge by default
	LPC_SC->EXTPOLAR &= ~(1 <<1);
	// Enable External Interrupt 1
	NVIC_EnableIRQ(EINT1_IRQn);
	NVIC_SetPriority(EINT1_IRQn, 0);

	// EINT2 Configurations
	//set the interrupt function to use the supplied callback
	// Set PINSEL4 [21:20] = 01 for P2.10 as EINT2
	LPC_PINCON->PINSEL4 &= ~(0b11 << 24);
	LPC_PINCON->PINSEL4 |= (0b01 << 24);
	//always clear the flag before initializing
	LPC_SC->EXTINT = EINT2_CLR;
	// EINT0 user selected trigger
	LPC_SC->EXTMODE |= EINT2_EDGE;
	// EINT2 is falling edge by default
	LPC_SC->EXTPOLAR &= ~(1 << 2);
	// Enable External Interrupt 2
	NVIC_EnableIRQ(EINT2_IRQn);
	NVIC_SetPriority(EINT2_IRQn, 0);

	// EINT3 Configurations
	//set the interrupt function to use the supplied callback
	// Set PINSEL4 [21:20] = 01 for P2.10 as EINT3
	LPC_PINCON->PINSEL4 &= ~(0b11 << 26);
	LPC_PINCON->PINSEL4 |= (0b01 << 26);
	//always clear the flag before initializing
	LPC_SC->EXTINT = EINT3_CLR;
	// EINT3 user selected trigger
	LPC_SC->EXTMODE |= EINT3_EDGE;
	// EINT3 is falling edge by default
	LPC_SC->EXTPOLAR &= ~(1 << 3);
	// Enable External Interrupt 3
	NVIC_EnableIRQ(EINT3_IRQn);
	NVIC_SetPriority(EINT3_IRQn, 0);

}

// Initializes PWM 1.5 on the board
void pwm_init(void) {

	// Select PCLK_PWM1 in the peripheral clock
	LPC_SC->PCLKSEL0 |= (PWM_CLKDIV << 12);

	// Select GPIO Mode to PWM
	LPC_PINCON->PINSEL4 |= (PWM_MODE << 8);

	// enable neither pull up nor pull down
	//LPC_PINCON->PINMODE4 |= 1 << 9;

	// Set Match Register to control PWM freq derived from prescaler
	LPC_PWM1->MR0 = MAXPWM;

	// Select Reset for Behavior on MR0 Counter Match to set the PWM freq
	LPC_PWM1->MCR |= (1 << 1);


	pwmduty = MINPWM;
	// Set Match Register to control duty cycle of PWM5.
	LPC_PWM1->MR5 = pwmduty;

	// Enable Timer and PWM
	LPC_PWM1->TCR |= (1 << 0) | (1 << 3);

	// Set PWM5 to single edge controlled mode
	//LPC_PWM1->PCR |= (0 << 5);

	// Enable PWM on P[2.4]
	LPC_PWM1->PCR |= (1 << 13);

}

// Used to calculate and display Hall Effect Data
void calc_rpm_data(void) {

    // Puts data from teeth counts to the DynoData struct.
	DynoData.rpm0 =  calc_count0;
	DynoData.rpm1 = calc_count1;
	DynoData.rpm2 = calc_count2;
	DynoData.rpm3 = calc_count3;

}

// Function to invert current state of leds
void leds_invert(void)
{
	// Read current state of GPIO P1_16..31, which includes LEDs
	//ledstate = LPC_GPIO2->FIOPIN;
	ledstate = LPC_GPIO1->FIOPIN;
	// Turn off LEDs that are on
	// (ANDing to ensure we only affect the LED outputs)
	//LPC_GPIO2->FIOCLR = ledstate & (1 << 13);
	LPC_GPIO1->FIOCLR = ledstate & (1 << 29);
	// Turn on LEDs that are off
	// (ANDing to ensure we only affect the LED outputs)
	//LPC_GPIO2->FIOSET = ((~ledstate) & (1 << 13));
	LPC_GPIO1->FIOSET = ((~ledstate) & (1 << 29));
}

// ***** EINT0 Interrupt Handler*****
void EINT0_IRQHandler(void)
{
	// Clear interrupt
	LPC_SC->EXTINT = EINT0_CLR;
	count0++;
}


// ***** EINT1 Interrupt Handler***** Flywheel
void EINT1_IRQHandler(void)
{
	// Clear interrupt
	LPC_SC->EXTINT = EINT1_CLR;
	count1++;
}

// ***** EINT2 Interrupt Handler*****
void EINT2_IRQHandler(void)
{
	// Clear interrupt
	LPC_SC->EXTINT = EINT2_CLR;
	count2++;
}

// ***** EINT3 Interrupt Handler*****
void EINT3_IRQHandler(void)
{
	// Clear interrupt
	LPC_SC->EXTINT = EINT3_CLR;
	count3++;
}

// Timer responsible for blinking MCU_LED with duty cycle .5 sec.
void TIMER0_IRQHandler(void)
{
	LPC_TIM0->IR = 1;			/* clear interrupt flag */
	//disable and reset the timer
	LPC_TIM0->TCR= 2;
	//restart the timer
	LPC_TIM0->TCR= 1;
	leds_invert();
}


void TIMER1_IRQHandler (void)
{
	LPC_TIM1->IR = 1;			/* clear interrupt flag */
	rpm_flag = 1;

	calc_count1 = count1;
	calc_count0 = count0;
	calc_count2 = count2;
	calc_count3 = count3;

	count1 = 0;
	count0 = 0;
	count2 = 0;
	count3 = 0;

}

// Responsible for sampling the channels of the external ADCs.
void sampleADC(void) {

	// Iterates through the various channels on the 1st ADC, sending and receiving data as long
	// as the bit for the particular channel is high.
	for(i=0;i<8;i++){
		if ((ADCStatusConv & ( 1 << i )) >> i) {
			SPISendN(send[i],2,1);
			SPIRecvN(rec[i],2,1);
			ADCdata[i] = ((rec[i][1] | (rec[i][0] << 8) )) & 0x0FFF;
		}
	}

	// Iterates through the various channels on the 2st ADC, sending and receiving data as long
	// as the bit for the particular channel is high.
	for(i=0;i<8;i++){
		if ((ADCStatusConv & ( 1 << (i+8) )) >> (i+8)) {
			SPISendN(send[i],2,2);
			SPIRecvN(rec[i],2,2);
			ADCdata[i+8] = ((rec[i][1] | (rec[i][0] << 8) )) & 0x0FFF;
		}
	}

    // Transfer data from ADCData to the DynoData struct.
	DynoData.adc0 = ADCdata[0];
	DynoData.adc1 = ADCdata[1];
	DynoData.adc2 = ADCdata[2];
	DynoData.adc3 = ADCdata[3];
	DynoData.adc4 = ADCdata[4];
	DynoData.adc5 = ADCdata[5];
	DynoData.adc6 = ADCdata[6];
	DynoData.adc7 = ADCdata[7];
	DynoData.adc8 = ADCdata[8];
	DynoData.adc9 = ADCdata[9];
	DynoData.adc10 = ADCdata[10];
	DynoData.adc11 = ADCdata[11];
	DynoData.adc12 = ADCdata[12];
	DynoData.adc13 = ADCdata[13];
	DynoData.adc14 = ADCdata[14];
	DynoData.adc15 = ADCdata[15];

}

// Parses the string from UART into usable integer values for the program.
void DAQ_Config(uint8_t* config_str) {

	// Parse Data String using strtok() into its constituent parts.
	uint8_t* parser;

    // This is for samples/second.
    parser = strtok(config_str,",");
	sprintf(samplePeriod,"%s",parser);

    // This is for parsing the string to retrieve EINT status.
	parser = strtok (NULL, ",");
	sprintf(eintStatus,"%s",parser);

    // This is for parsing the string to retrieve ADC channel status.
	parser = strtok (NULL, ",");
	sprintf(ADCStatus,"%s",parser);

	// Take parts from above and convert them from char arrays to ints.
	samplePeriodInt = atoi(samplePeriod);
	eintStatusConv = (int)strtol(eintStatus,NULL,0);
	ADCStatusConv = (int)strtol(ADCStatus,NULL,0);

}


// Responsible for parsing the data received via UART for the setting string.
void receiveConfigData(char c) {

	// If '$' has been detected, record characters received. Otherwise, dump them.
	if (configStringFlag == 1)
		rcvData[r++] = c;
	else
		r = 0;

	// if '$' detected, start recording the string.
	if (c == '$')
		configStringFlag = 1;

	// If '#' detected, stop recording the string and signal the start of data collection.
	if (c == '#') {
		configStringFlag = 0;
		configDataReady = 1;
		configStatus = 1;
		time = 0;
	}

	// If '&' detected, start recording data again and reset the timer.
	if (c == '&') {
		configStatus = 1;
		time = 0;
	}


	// If '@' detected, stop recording data.
	if (c == '@') {
		configStatus = 0;
	}

}

void asciiSensorStatus(uint16_t eintStatusConv, uint32_t ADCStatusConv, uint8_t* datastr) {

	memset(datastr,0,90); // Used to clear the datastr buffer.

	uint32_t q; // Variable used for iteration.
	uint8_t buf[20]; // Buffer that is concatenated onto the datastr buffer.

	// Creating an array so that the values of the struct can be iterated through.
	uint32_t data[20] = {DynoData.rpm0, DynoData.rpm1, DynoData.rpm2, DynoData.rpm3,
			DynoData.adc0, DynoData.adc1, DynoData.adc2, DynoData.adc3, DynoData.adc4,
			DynoData.adc5, DynoData.adc6, DynoData.adc7, DynoData.adc8, DynoData.adc9,
			DynoData.adc10, DynoData.adc11, DynoData.adc12, DynoData.adc13,
			DynoData.adc14, DynoData.adc15};

	// Always concatenate the time value into datastr.
	time = time + samplePeriodInt;
	sprintf(buf,"%u ", time);
	strcat(datastr,buf);


	// Check the eintStatusConv integer bit by bit to see if a particular EINT is enabled.
	// If so, print its results in the datastr buffer.
	for(q = 0; q < 4; q++) {

		if((eintStatusConv & ( 1 << q )) >> q) {
			sprintf(buf,"%u ", data[q]);
			strcat(datastr,buf);
		}
	}

	// Check the ADCStatusConv integer bit by bit to see if a particular EINT is enabled.
	// If so, print its results in the datastr buffer.
	for(q = 0; q < 16; q++) {

		if((ADCStatusConv & ( 1 << q )) >> q) {
			sprintf(buf,"%u ", data[q+4]);
			strcat(datastr,buf);
		}
	}

	// calculate the checksum for the string
	for(q = 0; q < sizeof(datastr); q++){
		checksum = checksum ^ datastr[q];
	}

	checksumByteOne = checksum >> 4;
	checksumByteOne |= (1<<6);

	checksumByteTwo = 0x0F & checksum;
	checksumByteTwo |= (1<<6);


	sprintf(buf,"%c%c", checksumByteOne, checksumByteTwo);
	strcat(datastr,buf);

	// Need this for the data to advance to the next line.
	strcat(datastr,"\r\n");

}


int main(void) {


	// Initialize Timers
	timer_init();

	// Initialize External Interrupts
	eint_init();

	// Initialize PWM - used if control of a servo is desired. 
	//pwm_init();

	// Initialize to receive the setting string via UART.
	r = 0;

	// Initialize UART Communications OPERATES AT 115200
	UART2_Init(115200, receiveConfigData);

	// Initialize LED
	// Set P1_29 to 00 - GPIO
	LPC_PINCON->PINSEL3	&= (~0x0C000000);
	// Set GPIO - P1_29 - to be an output
	//LPC_GPIO2->FIODIR |= (1 << 13);
	LPC_GPIO1->FIODIR |= (1 << 29);

	// Set the LED to be on.
	LPC_GPIO1->FIOSET |= (1 << 29);

    // Initialize SPI.
	SPIInit();

	while(1) {

		// If the config string has been received, allow the DAQ to print the data to UART.
		if(configStatus == 1) { 

			// If a new config string has been received,
			// update the integers governing DAQ behavior.
			if(configDataReady == 1){
				// Parse the new config string into the appropriate configuration integers.
				DAQ_Config(rcvData);
				//disable and reset the timer
				LPC_TIM1->TCR= 2;
				// Calculate the new timer match register value.
				LPC_TIM1->MR0 = (int)(33.33*samplePeriodInt);
				//restart the timer
				LPC_TIM1->TCR= 1;
				configDataReady = 0;
			}

			// If the time between samples has passed, send a new line of data via UART.
			if (rpm_flag == 1) {
				calc_rpm_data();	// Calculate the number of teeth seen since the last output.
				rpm_flag = 0;		// Reset rpm_flag.
				sampleADC();		// Sample the channels of the ADCs.
				// Print the data to the datastr buffer.
				asciiSensorStatus(eintStatusConv, ADCStatusConv, datastr);
				UART2_PrintString(datastr); // Send the data via UART2.
				// Wait until UART2 is done sending the data.
				// (This prevents errors in the UART output.)
				while(UART2_is_not_Ready());

			}

		}

	}

	return 0;

}




