/*
 * This is example code to set up uart2 (works for uart0 and 3 with small
 * modifications).
 *
 * Thanks to Brian Toth of Cornell Baja SAE.
 */

// Code used in Michael Kilzer's (mjk277) Cornell University MEng Project.

#include "LPC17xx.h"
#include "uart2.h"
#include "definitions.h"

// PCUART2
#define PCUART2_POWERON (1 << 24)

#define PCLK_UART2 16
#define PCLK_UART2_MASK (3 << 16)

#define IER_RBR		0x01
#define IER_THRE	0x02
#define IER_RLS		0x04

#define IIR_RLS		0x06
#define IIR_RDA		0x04
#define IIR_CTI		0x0C
#define IIR_THRE	0x02
#define IIR_ABEO	0x100
#define IIR_ABTO	0x200
#define INTERRUPT_SOURCE_MASK 0xF

#define LSR_RDR		(1 << 0)
#define LSR_OE		(1 << 1)
#define LSR_PE		(1 << 2)
#define LSR_FE		(1 << 3)
#define LSR_BI		(1 << 4)
#define LSR_THRE	(1 << 5)
#define LSR_TEMT	(1 << 6)
#define LSR_RXFE	(1 << 7)
#define ERROR_CONDITION_MASK 0x1E

#define TX_FIFO_SIZE 16 //determined through experimentation
#define RX_FIFO_SIZE 14 //set with FCR

static receive_func data_receive2;

static uint8_t txbuf2[TX_SIZE2];

static Uart_buf b2;

// ***********************
// Function to set up UART
void UART2_Init(int baudrate, receive_func interrupt_function){
	int pclk;
	unsigned long int Fdiv;

	//set the interrupt to use the user-specified function
	data_receive2= interrupt_function;

	//set uart2 tx struct
	b2.read= 0;
	b2.write= 0;
	b2.x= txbuf2;

	// PCLK_UART2 is being set to 1/4 of SystemCoreClock
	pclk = SystemCoreClock / 4;	
	
	// Turn on power to UART2
	LPC_SC->PCONP |=  PCUART2_POWERON;
		
	// Turn on UART2 peripheral clock
	LPC_SC->PCLKSEL1 &= ~(PCLK_UART2_MASK);
	LPC_SC->PCLKSEL1 |=  (0 << PCLK_UART2);		// PCLK_periph = CCLK/2
	
	// Set PINSEL0 so that P2.8 = TXD2, P2.9 = RXD2
	LPC_PINCON->PINSEL4 &= ~0x0f0000;
	LPC_PINCON->PINSEL4 |= ((1 << 17) | (1 << 19));
	// Set PINMODE0 so that TXD2 and RXD2 have neither pull-up nor pull-down resistors enabled.
	//LPC_PINCON->PINMODE4 |= ((1 << 17) | (1 << 19));
	
	LPC_UART2->LCR = 0x83;		// 8 bits, no Parity, 1 Stop bit, DLAB=1
    Fdiv = ( pclk / 16 ) / baudrate ;	// Set baud rate

    LPC_UART2->FDR = 0x21;
    LPC_UART2->DLM = 0;
    LPC_UART2->DLL = 3;
    //LPC_UART2->DLM = Fdiv / 256;
   // LPC_UART2->DLL = Fdiv % 256;


    LPC_UART2->LCR = 0x03;		// 8 bits, no Parity, 1 Stop bit DLAB = 0
    LPC_UART2->FCR = 0xC7;		// Enable and reset TX and RX FIFO
    							// FIFO queues up X bytes received before
    							// triggering an interrupt
    							// note that the TX FIFO is always 16 bytes
    							// 14 bytes RX FIFO = 0xC7

    LPC_UART2->IER = IER_RBR | IER_THRE;
    						  //enable Receive Data Enable Interrupt
    					   	  //and enable the Transmit Holding Register
    					  	  //Empty interrupt

    //line status interrupt could be used for error checking.


    //should probably go outside of this routine (so interrupts don't happen during
    // other initializations)
	NVIC_EnableIRQ(UART2_IRQn);
}

// ***********************
// Demonstration function to send character over UART
int8_t UART2_Sendchar(char c)
{
	//check to see if the tx is already "primed" (that is to say there is at least
	//one character being sent
	if(LPC_UART2->LSR & LSR_TEMT){
		//we need to "prime" the tx by throwing a character in there.  can't just disable and
		//enable the tx interrupt like on a normal microcontroller
		LPC_UART2->THR= c;
	} else {
		if(b2.write >= TX_SIZE2){
			//we have overflowed the buffer
			return -1;
		}
		b2.x[b2.write]= c;
		b2.write++;
	}
	return 0;
}

// ***********************
// Function to prints the string out over the UART
//prints number of characters sent
int8_t UART2_PrintString(char *pcString){
	int i = 0;
	// loop through until reach string's zero terminator
	while (pcString[i] != 0) {	
		if(UART2_Sendchar(pcString[i])){ // print each character
			return i;
		}
		i++;
	}
	return i;
}

// ***********************
// Function to prints the string out over the UART
//prints number of characters sent
int8_t UART2_SendEncoded(char *pcString, int lengthData){
	int j;
	// loop through until reach string's zero terminator
	for (j=0; j<lengthData; j++) {
		if(UART2_Sendchar(pcString[j])){ // print each character
			return j;
		}
		j++;
	}
	return j;
}

//if you want to block
int8_t UART2_is_not_Ready(){
	if(b2.write == 0){
		return 0;
	} else {
		return 1;
	}
}

static inline void read_full_rx_fifo(void){
	char i, line_status;

	//as long as there is data to read, read it.
	//note that there will be at least one character
	//available since the interrupt fired.
	line_status = LPC_UART2->LSR;

	while(line_status & LSR_RDR){

		i= LPC_UART2->RBR;

		//if there is an error, handle it appropriately
		if(line_status & LSR_RXFE){
			//throw away the character with an error

			//handle the error somehow
			switch(line_status & ERROR_CONDITION_MASK){
			//parity error
			case LSR_PE:
				break;
				//Framing error
			case LSR_FE:
				break;
				//Break interrupt
			case LSR_BI:
				break;
			}
		} else {
			//add character to rx buffer.  if the buffer is full,
			//this character will just be deleted.

			//currently having a problem in which the buffer gets in a bad state and
			//never escapes.  fixing this by have it read a large chunk from the buffer.

			//i believe i have fixed the above
			data_receive2(i);
		}

		//check for the next character
		line_status = LPC_UART2->LSR;
	}
}

static inline void fill_tx_fifo(void){
	uint8_t i= 0;
	//write as much of we can of what we have to the FIFO buffer
	while(b2.write - b2.read > 0 && i < TX_FIFO_SIZE){
		LPC_UART2->THR = b2.x[b2.read];
		i++;
		b2.read++;
	}

	//reset the sending buffer if possible
	if(b2.write == b2.read){
		b2.write= 0;
		b2.read= 0;
	}
}

// ***********************
// UART Interrupt Handler
void UART2_IRQHandler (void)
{
	//read the interrupt status
	char check = LPC_UART2 -> IIR & INTERRUPT_SOURCE_MASK;
	char error_cond;

	switch(check){
	//receive line status (error detected)
	case IIR_RLS:
		//note that PE, FE, and BI each correspond to a single
		//byte.  to find the byte this error refers to, you must
		//first read the LSR, and then read a byte from the RBR
		error_cond= LPC_UART2->LSR & ERROR_CONDITION_MASK;
		switch(error_cond){
		//overrun error
		case LSR_OE:
			break;

		//parity error
		case LSR_PE:
			break;

		//Framing error
		case LSR_FE:
			break;

		//Break interrupt
		case LSR_BI:
			break;

		default:
			break;
		}
		break;

	//receive data available (rx)
	case IIR_RDA:
		read_full_rx_fifo();
		break;

	//character time-out (received fewer than 14 bytes)
	case IIR_CTI:
		//uses the same code as above right now
		read_full_rx_fifo();
		break;


	//transmit holding register empty
	case IIR_THRE:
		fill_tx_fifo();
		break;

	default:
		break;
	}

	//auto-baud process complete
	if(LPC_UART2->IIR & IIR_ABEO){

	}

	//auto-baud process timeout
	if(LPC_UART2->IIR & IIR_ABTO){

	}

	return;
}

